import ipaddress
import socket
import json
from . import ovsdb_query
from .bridge import OvsBridge
from .port import OvsPort
from datetime import datetime, timedelta
from . import ovspy_error
import sys
import time

class OvsClient:
    SEND_DEBUG = False
    RECV_DEBUG = False
    
    def __init__(self, ovsdb_port, ovsdb_ip="127.0.0.1"):
        self._ovsdb_ip = ipaddress.ip_address(ovsdb_ip)
        self._ovsdb_port = int(ovsdb_port)
        self._query_timeout = 5
    
    def _send(self, query):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((str(self._ovsdb_ip), self._ovsdb_port))
        
        if self.SEND_DEBUG:
            sys.stderr.write("[SEND] %s\n" % json.dumps(query).encode())
        s.send(json.dumps(query).encode())
        #s.shutdown(socket.SHUT_RDWR)
        
        buf = bytes()
        bufsize = 16
        
        timeout = datetime.now() + timedelta(seconds=self._query_timeout)
        while True:
            if datetime.now() >= timeout:
                raise ovspy_error.TransactionError("Timeout")
            
            buf += s.recv(bufsize)
            
            try:
                query_result = json.loads(buf.decode())
                
                #echo method
                #https://tools.ietf.org/html/rfc7047
                if "method" in query_result.keys() and query_result["method"] == "echo":
                    echo_reply= {
                        "method": "echo",
                        "params": query_result["params"],
                        "id": query_result["id"]
                    }
                    s.send(json.loads(echo_reply).encode())
                    buf = bytes()
                    continue
                else:
                    break
            except json.JSONDecodeError:
                pass
        
        s.close()
        
        if self.RECV_DEBUG:
            sys.stderr.write("[RECV] %s\n" % query_result)
        
        self._check_error(query_result)
        return query_result
    
    @staticmethod
    def _check_error(query_result_json):
        if "result" in query_result_json.keys():
            for item in query_result_json["result"]:
                if "error" in item.keys():
                    raise ovspy_error.TransactionError("[QueryError] %s" % item["details"])
        elif len(query_result_json["error"]) != 0:
            raise ovspy_error.TransactionError("[QueryError] %s" % query_result_json["error"])
    
    #get Open_vSwitch table
    def get_ovs_raw(self):
        query = ovsdb_query.Generator.get_ovs()
        result = self._send(query)
        return result
    
    #get id of Open_vSwitch entry from Open_vSwitch table
    def get_uuid(self):
        return self.get_ovs_raw()["result"][0]["rows"][0]["_uuid"][1]
    
    def get_bridge_raw(self, bridge_id=None):
        query = ovsdb_query.Generator.get_bridges()
        
        result = self._send(query)
        
        if bridge_id is None:
            return result["result"][0]["rows"]
        else:
            for br in result["result"][0]["rows"]:
                if br['_uuid'][1] == bridge_id:
                    return br
        
        return None
    
    def get_bridges(self):
        bridges = self.get_bridge_raw()
        ret = []
        
        for br in bridges:
            _br = OvsBridge(br['_uuid'][1])
            _br.set_client(self)
            ret.append(_br)
        
        return ret
    
    def find_bridge(self, bridge_name):
        for br in self.get_bridges():
            if br.get_name() == bridge_name:
                return br
        return None
    
    def find_port(self, port_name):
        for p in self.get_port_raw():
            if p["name"] == port_name:
                return p
        return None
    
    def get_port_raw(self, port_id=None):
        
        query = ovsdb_query.Generator.get_ports()
        
        result = self._send(query)
        
        if port_id is not None:
            for p in result["result"][0]["rows"]:
                if p['_uuid'][1] == port_id:
                    return p
        else:
            return result["result"][0]["rows"]
        
        return None
    
    def add_port_to_bridge(self, bridge, port_name, vlan=None):
        bridge_raw = bridge.get_raw()
        if bridge_raw is None:
            raise ovspy_error.NotFound("bridge is not found")
        
        if self.find_port(port_name) is not None:
            raise ovspy_error.Duplicate("port is already exist")
        
        #print(bridge.get_raw())
        
        exist_ports = []
        for p in bridge.get_ports():
            exist_ports.append(p.get_uuid())
        
        query = ovsdb_query.Generator.add_port(bridge.get_uuid(), exist_ports, port_name, vlan=vlan)
        self._send(query)
    
    def del_port_from_bridge(self, bridge, port_name):
        target_port = bridge.find_port(port_name)
        
        exist_ports = []
        for p in bridge.get_ports():
            exist_ports.append(p.get_uuid())
        exist_ports = list(set(exist_ports))
        
        if target_port is None:
            raise ovspy_error.NotFound("Specified port(%s) is not exist in bridge(%s)." % (port_name, bridge.get_name()))
        if target_port.get_uuid() not in exist_ports:
            raise ovspy_error.NotFound("Specified port(%s) is not exist in bridge(%s)." % (port_name, bridge.get_name()))
        
        query = ovsdb_query.Generator.del_port(bridge.get_uuid(), exist_ports, target_port.get_uuid())
        self._send(query)
    
    def add_bridge(self, bridge_name):
        exist_bridges = []
        for br in self.get_bridges():
            if bridge_name == br.get_name():
                raise ovspy_error.Duplicate("Bridge(%s) is already exist." % bridge_name)
            exist_bridges.append(br.get_uuid())
            
        exist_bridges = list(set(exist_bridges))
        
        query = ovsdb_query.Generator.add_bridge(self.get_uuid(), bridge_name, exist_bridges)
        self._send(query)
        
    def del_bridge(self, bridge_name):
        target_bridge = self.find_bridge(bridge_name)
        
        exist_bridges = []
        for br in self.get_bridges():
            exist_bridges.append(br.get_uuid())
        
        if target_bridge is None:
            raise ovspy_error.NotFound("Bridge(%s) is not exist." % bridge_name)
        if target_bridge.get_uuid() not in exist_bridges:
            raise ovspy_error.NotFound("Bridge(%s) is not exist." % bridge_name)
        
        query = ovsdb_query.Generator.del_bridge(self.get_uuid(), exist_bridges, target_bridge.get_uuid())
        self._send(query)
    




class OvspyError(Exception):
    pass

class TransactionError(OvspyError):
    pass

class Duplicate(OvspyError):
    pass

class NotFound(OvspyError):
    pass

class Unsupported(OvspyError):
    pass



import copy

class OvsPort():
    def __init__(self, uuid):
        self.id = uuid
    
    def set_client(self, client):
        self.ovs_client = client
    
    def get_uuid(self):
        return self.id
    
    #def get_info(self):
    #    return copy.deepcopy(self.ovs_client.get_port_raw(port_id=self.id))
    
    def get_name(self):
        target_port = self.ovs_client.get_port_raw(port_id=self.id)
        return target_port["name"]
    
    def get_vlan_info(self):
        target_port = self.ovs_client.get_port_raw(port_id=self.id)
        
        if isinstance(target_port["tag"], list):
            return {
                "mode":"trunk",
                "tag": target_port["trunks"][1]
            }
        elif isinstance(target_port["tag"], int):
            return {
                "mode":"access",
                "tag":target_port["tag"]
            }
        
        raise ovspy_error.Unsupported("The VLAN configuration parameters is unsupported ovspy. (%s)", target_port)
    
    def get_raw(self):
        return self.ovs_client.get_port_raw(port_id=self.get_uuid())
    




from .port import OvsPort

class OvsBridge():
    def __init__(self, uuid):
        self.id = uuid
        
    def set_client(self, client):
        self.ovs_client = client
    
    def get_uuid(self):
        return self.id
    
    def get_ports(self):
        target_bridge = self.ovs_client.get_bridge_raw(bridge_id=self.id)
        
        ret = []
        #If ports is exist at the bridge, 'set' is store at ["port"][0]
        #'set' is not exist, the bridges doesn't have ports
        if target_bridge["ports"][0] == 'set':
            for port in target_bridge["ports"][1]:
                _port = OvsPort(port[1])
                _port.set_client(self.ovs_client)
                ret.append(_port)
        elif target_bridge["ports"][0] == "uuid":
            _port = OvsPort(target_bridge["ports"][1])
            _port.set_client(self.ovs_client)
            ret.append(_port)
        
        return ret
    
    def find_port(self, port_name):
        for p in self.get_ports():
            if p.get_name() == port_name:
                return p
        return None
    
    def get_raw(self):
        return self.ovs_client.get_bridge_raw(bridge_id=self.id)
    
    def get_name(self):
        target_bridge = self.ovs_client.get_bridge_raw(bridge_id=self.id)
        return target_bridge['name']
    
    def add_port(self, port_name, vlan=None):
        self.ovs_client.add_port_to_bridge(self, port_name, vlan)
    
    def del_port(self, port_name, vlan=None):
        self.ovs_client.del_port_from_bridge(self, port_name)
    



import uuid
import random

class Generator():
    
    id_min = 0
    id_max = 10000
    
    @staticmethod
    def get_ovs():
        query = {
            "method":"transact",
            "params":[
                "Open_vSwitch",
                {
                    "op" : "select",
                    "table" : "Open_vSwitch",
                    "where" : [],
                }
            ],
            "id": random.randint(Generator.id_min, Generator.id_max)
        }
        
        return query
    
    @staticmethod
    def get_bridges():
        query = {
            "method":"transact",
            "params":[
                "Open_vSwitch",
                {
                    "op" : "select",
                    "table" : "Bridge",
                    "where" : [],
                }
            ],
            "id": random.randint(Generator.id_min, Generator.id_max)
        }
        
        return query
    
    @staticmethod
    def get_ports():
        query = {
            "method":"transact",
            "params":[
                "Open_vSwitch",
                {
                    "op" : "select",
                    "table" : "Port",
                    "where" : [],
                }
            ],
            "id": random.randint(Generator.id_min, Generator.id_max)
        }
        
        return query
    
    @staticmethod
    def add_port(bridge_id, exist_port_id_list, new_port_name, vlan=None):
        
        #generate temporary values for query string needs
        interface_tmp_id = "row%s" % str(uuid.uuid4()).replace("-", "_")
        port_tmp_id = "row%s" % str(uuid.uuid4()).replace("-", "_")
        
        #generate new port list
        ports = []
        for p_id in exist_port_id_list:
            ports.append(["uuid", p_id])
        ports.append(["named-uuid", port_tmp_id])
        
        query = {
            "method":"transact",
            "params":[
                "Open_vSwitch",
                {
                    "uuid-name" : interface_tmp_id,
                    "op" : "insert",
                    "table" : "Interface",
                    "row":{
                        "name":new_port_name,
                        #"type":""
                    }
                },
                {
                    "uuid-name": port_tmp_id,
                    "op" : "insert",
                    "table" : "Port",
                    "row":{
                        "name" : new_port_name,
                        "interfaces":[
                            "named-uuid",
                            interface_tmp_id
                        ]
                    }
                },
                {
                    "where": [
                        [
                            "_uuid",
                            "==",
                            [
                                "uuid",
                                bridge_id
                            ]
                        ]
                    ],
                    "row": {
                        "ports": [
                            "set",
                            ports
                        ]
                    },
                    "op": "update",
                    "table": "Bridge"
                }
            ],
            "id": random.randint(Generator.id_min, Generator.id_max)
        }
        
        if vlan is not None:
            if isinstance(vlan, int):
                query["params"][2]["row"].update({"tag":vlan})
            elif isinstance(vlan, list):
                query["params"][2]["row"].update({"tag":["set", []], "trunks":["set", vlan]})
            else:
                raise ValueError("Invalid VLAN type or format was specified.")
        return query
    
    @staticmethod
    def del_port(bridge_id, bridge_port_id_list, target_port_id):
        bridge_port_id_list.remove(target_port_id)
        new_ports = []
        for p in bridge_port_id_list:
            new_ports.append(["uuid", p])
        
        ret = {
            "id": random.randint(Generator.id_min, Generator.id_max),
            "method": "transact",
            "params": [
                "Open_vSwitch",
                {
                    "where": [
                        [
                            "_uuid",
                            "==",
                            [
                                "uuid",
                                bridge_id
                            ]
                        ]
                    ],
                    "row": {
                        "ports": [
                            "set",
                            new_ports
                        ]
                    },
                    "op": "update",
                    "table": "Bridge"
                }
            ]
        }
        
        return ret
    
    @staticmethod
    def add_bridge(ovs_id, bridge_name, exist_bridge_id_list):
        #generate temporary values for query string needs
        bridge_tmp_id = "row%s" % str(uuid.uuid4()).replace("-", "_")
        inital_port_tmp_id = "row%s" % str(uuid.uuid4()).replace("-", "_")
        inital_interface_tmp_id = "row%s" % str(uuid.uuid4()).replace("-", "_")
        
        bridges = []
        for br_id in exist_bridge_id_list:
            bridges.append(["uuid", br_id])
        bridges.append(["named-uuid", bridge_tmp_id])
        
        ret = {
            "id": random.randint(Generator.id_min, Generator.id_max),
            "method": "transact",
            "params": [
                "Open_vSwitch",
                {
                    "where": [
                        [
                            "_uuid",
                            "==",
                            [
                                "uuid",
                                ovs_id
                            ]
                        ]
                    ],
                    "row": {
                        "bridges": [
                            "set",
                            bridges
                        ]
                    },
                    "op": "update",
                    "table": "Open_vSwitch"
                },
                {
                    "uuid-name": inital_port_tmp_id,
                    "row": {
                        "name": bridge_name,
                        "interfaces": [
                            "named-uuid",
                            inital_interface_tmp_id
                        ]
                    },
                    "op": "insert",
                    "table": "Port"
                },
                {
                    "uuid-name": bridge_tmp_id,
                    "row": {
                        "name": bridge_name,
                        "ports": [
                            "named-uuid",
                            inital_port_tmp_id
                        ]
                    },
                    "op": "insert",
                    "table": "Bridge"
                },
                {
                    "uuid-name": inital_interface_tmp_id,
                    "row": {
                        "name": bridge_name,
                        "type": "internal"
                    },
                    "op": "insert",
                    "table": "Interface"
                },
            ]
        }
        
        return ret
    
    @staticmethod
    def del_bridge(ovs_id, bridge_id_list, target_bridge_id):
        bridge_id_list.remove(target_bridge_id)
        new_bridges = []
        for br in bridge_id_list:
            new_bridges.append(["uuid", br])
        
        ret = {
            "id": random.randint(Generator.id_min, Generator.id_max),
            "method": "transact",
            "params": [
                "Open_vSwitch",
                {
                    "where": [
                        [
                            "_uuid",
                            "==",
                            [
                                "uuid",
                                ovs_id
                            ]
                        ]
                    ],
                    "row": {
                        "bridges": [
                            "set",
                            new_bridges
                        ]
                    },
                    "op": "update",
                    "table": "Open_vSwitch"
                }
            ]
        }
        
        return ret



from setuptools import setup, find_packages

requires = []

setup(
    name='ovspy',
    version='0.1.0b',
    description='Open vSwitch Mnipulation Library',
    url='https://github.com/kamaboko123/ovspy',
    author='kamaboko123',
    author_email='6112062+kamaboko123@users.noreply.github.com',
    license='MIT',
    packages=['ovspy'],
    install_requires=requires,
    classifiers=[
        'Programming Language :: Python :: 3.5',
    ]
)




class Generator():
    
    @staticmethod
    def get_bridges():
        query = {
            "method":"transact",
            "params":[
                "Open_vSwitch",
                {
                    "op" : "select",
                    "table" : "Bridge",
                    "where" : [],
                }
            ],
            "id":0
        }
        
        return query


